@@ -93,9 +93,16 @@ fn print_help() {
 <Enter>    Print current byte(s) and move forward to next set of byte(s)
 314        Move to byte number 0x314 (or 0d314 depending on 'x') and print from
                there
++          Move 1 byte forward and print from there
+-          Move 1 byte back and print from there
++314       Move 0x314 bytes forward (or 0d314 depending on 'x') and print from
+               there
+-314       Move 0x314 bytes back (or 0d314 depending on 'x') and print from
+               there
 $          Move to last byte and print it
-12,34p     Print bytes 12 - 34 inclusive, then move to byte 0x34
-               (or 0d34 depending on 'x')
+i          Prompt you to write out bytes which will be inserted at current index
+12,34p     Print bytes 12 - 34 inclusive (depending on 'x'), then move to
+                leftmost byte printed on the last line.
 n          Toggle whether or not byte numbers are printed before bytes
 p          Print current byte(s) (depending on 'W')
 s          Print state of all toggles and 'W'idth
@@ -107,6 +114,29 @@ q          quit
 }
 
 
+fn read_bytes_from_user() -> Result<Vec<u8>, String> {
+    print!("> ");
+    io::stdout().flush().unwrap();
+    let input = match get_input_or_die() {
+        Ok(input) => input,
+        Err(_errcode) => {
+            return Err("Couldn't read input".to_owned());
+        }
+    };
+
+    // TODO Allow general whitespace, not just literal spaces
+    let re_bytes = Regex::new(r"^ *([0-9a-fA-F][0-9a-fA-F] *)* *$").unwrap();
+    if re_bytes.is_match(&input) {
+        let nibbles:Vec<String> = input.replace(" ", "").chars().map(|x| x.to_string()).collect();
+        Ok(nibbles.chunks(2).map(|x| x.join("")).map(|x| u8::from_str_radix(&x, 16).unwrap()).collect())
+
+    }
+    else {
+        Err(format!("Couldn't interpret '{}' as a sequence of bytes", &input))
+    }
+}
+
+
 fn bad_range(bytes: &Vec<u8>, range: (usize, usize)) -> bool {
     bytes.len() == 0 || range.1 >= bytes.len()
 }
@@ -126,7 +156,7 @@ impl Command {
         let re_blank_line = Regex::new(r"^ *$").unwrap();
         let re_plus = Regex::new(r"^ *\+ *$").unwrap();
         let re_minus = Regex::new(r"^ *\- *$").unwrap();
-        let re_single_char_command = Regex::new(r"^ *(?P<command>[?npsxq]).*$").unwrap();
+        let re_single_char_command = Regex::new(r"^ *(?P<command>[?npsxqi]).*$").unwrap();
         let re_range = Regex::new(r"^ *(?P<begin>[0-9a-fA-F.$]+) *, *(?P<end>[0-9a-fA-F.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_specified_index = Regex::new(r"^ *(?P<index>[0-9A-Fa-f.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_offset_index = Regex::new(r"^ *(?P<sign>[-+])(?P<offset>[0-9A-Fa-f]+) *(?P<the_rest>.*) *$").unwrap();
@@ -558,18 +588,20 @@ mod tests {
 
 fn print_state(state:&State) {
     println!("At byte {} of {}", lino(&state),
-            hex_unless_dec(state.all_bytes.len(), state.radix));
+            hex_unless_dec_with_radix(state.all_bytes.len(), state.radix));
     if state.show_byte_numbers {
         println!("Printing byte numbers in {}",
             string_from_radix(state.radix));
     };
     println!("Interpreting input numbers as {}",
             string_from_radix(state.radix));
-    if state.radix == 10 {
-        println!("Printing a newline every 0d{} bytes", state.width);
+    println!("Printing a newline every {} bytes",
+            hex_unless_dec_with_radix(usize::from(state.width), state.radix));
+    if state.unsaved_changes {
+        println!("Unwritten changes");
     }
     else {
-        println!("Printing a newline every 0x{:x} bytes", state.width);
+        println!("No unwritten changes");
     }
 }
 
@@ -624,6 +656,18 @@ fn number_dot_dollar(index:usize, _max_index:usize, input:&str, radix:u32)
 }
 
 
+fn hex_unless_dec_with_radix(number:usize, radix:u32) -> String {
+    let letter = if radix == 10 {
+        'd'
+    }
+    else {
+        'x'
+    };
+
+    format!("0{}{}", letter, hex_unless_dec(number, radix))
+}
+
+
 fn hex_unless_dec(number:usize, radix:u32) -> String {
     if radix == 10 {
         format!("{}", number)
@@ -635,13 +679,14 @@ fn hex_unless_dec(number:usize, radix:u32) -> String {
 
 
 fn lino(state:&State) -> String {
-    hex_unless_dec(state.index, state.radix)
+    hex_unless_dec_with_radix(state.index, state.radix)
 }
 
 
 struct State {
     radix: u32,
     show_byte_numbers: bool,
+    unsaved_changes: bool,
 
     /* Current byte number, 0 to len -1 */
     index: usize,
@@ -658,9 +703,9 @@ struct State {
 
 impl fmt::Debug for State {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "radix: {}|show_byte_numbers: {}|index: {}|width: {}|n_padding: '{}'|",
-                self.radix, self.show_byte_numbers, self.index, self.width,
-                self.n_padding)
+        write!(f, "radix: {}|unsaved_changes: {}|show_byte_numbers: {}|index: {}|width: {}|n_padding: '{}'|",
+                self.radix, self.unsaved_changes, self.show_byte_numbers,
+                self.index, self.width, self.n_padding)
     }
 }
 
@@ -716,6 +761,7 @@ pub fn actual_runtime(filename: &str) -> i32 {
     let mut state = State{
         radix: 16,
         show_byte_numbers: true,
+        unsaved_changes: false,
         index: 0,
         width: NonZeroUsize::new(16).unwrap(),
         all_bytes: all_bytes,
@@ -785,6 +831,32 @@ pub fn actual_runtime(filename: &str) -> i32 {
                     }
                 },
 
+                /* insert */
+                'i' => {
+                    match read_bytes_from_user() {
+                        Ok(entered_bytes) => {
+                            // TODO Find the cheapest way to do this (maybe
+                            // make state.all_bytes a better container)
+                            let mut new = Vec::with_capacity(state.all_bytes.len() + entered_bytes.len());
+                            for i in 0..state.index {
+                                new.push(state.all_bytes[i]);
+                            }
+                            // TODO Could use Vec::splice here
+                            for i in 0..entered_bytes.len() {
+                                new.push(entered_bytes[i]);
+                            }
+                            for i in state.index..state.all_bytes.len() {
+                                new.push(state.all_bytes[i]);
+                            }
+                            state.all_bytes = new;
+                            state.unsaved_changes = true;
+                        },
+                        Err(error) => {
+                            println!("? ({})", error);
+                        },
+                    }
+                },
+
                 /* Help */
                 '?'|'h' => {
                     print_help();
@@ -873,7 +945,7 @@ pub fn actual_runtime(filename: &str) -> i32 {
 
                 /* Catchall error */
                 _ => {
-                    println!("? (Don't understand {})", command.command);
+                    println!("? (Don't understand command '{}')", command.command);
                     continue;
                 },
             }
